# -*- coding: utf-8 -*-
"""
Created on 2020.03.02

@author: MiniUFO
Copyright 2018. All rights reserved. Use is subject to license terms.
"""
import os, sys, re
import numpy as np
from datetime import datetime
from numpy import datetime64, timedelta64


"""
Core classes are defined below
"""
class CtlDescriptor(object):
    """A class for a ctl file
    
    This class represents a descriptor file like the .ctl file for GrADS.
    It generally includes a multi-dimensional dataset on the Earth.
    
    Attributes
    ----------
    dsetPath: str
        dataset file
    descPath: str
        descriptor file
    indxPath: str
        index file (for GRIB file)
    stnmPath: str
        station map file (for station file)
    
    pdef: PDEF
        Projection-definition
    tdef: Coordinate
        time coordinate definition
    zdef: Coordinate
        z-coordinate definition
    ydef: Coordinate
        y-coordinate definition
    xdef: Coordinate
        x-coordinate definition
    vdef: list of CtlVar
        Variables definition
    edef: list of str
        Ensemble definition
    
    comments: list of str
        list of global string comments
    
    zrev: bool
        z-dimension reverse (i.e., from north to south)
    yrev: bool
        y-dimension reverse (i.e., from upper to lower levels)
    
    hasData: bool
        whether the corresponding binary data file exist
    vcount: int
        variable count
    
    dtype: numpy.dtype
        data type
    periodicX: bool
        whether xdef is periodic
    
    cal365Days: bool
        whether the calendar is always 365 days (no leap year)
    template: bool
        whether it is a template for multiple binary files
    sequential: bool
        whether it is a sequential file (Fortran style)
    byteOrder: str
        byte order, ['little', 'big'] for little-endian or big-endian
    storage: str
        storage type, '99' or '-1,20' or others
    
    totalZCount: int
        total number of horizontal slice
    zRecLength: int
        record length of a single horizontal slice
    tRecLength: int
        record length of a single time (including all variables)
    """
    def __init__(self, encoding='GBK', **kwargs):
        """Constructor

        One of the keyword argument `file` or `content` should be specified.
        
        Parameters
        ----------
        encoding: {'GBK', 'UTF-8'}, optional
            Encoding for the ctl file contents.
        file: str
            The ctl path/file name.
        content: str
            A string representation for the ctl file contents.

        Returns
        -------
        CtlDescriptor
            An object represents the ctl file
        """
        self.vcount = 0
        
        self.pdef = None
        self.tdef = None
        self.zdef = None
        self.ydef = None
        self.xdef = None
        self.vdef = None
        self.edef = None
        self.comments = {}
        
        self.dsetPath = ''
        self.descPath = ''
        self.indxPath = ''
        self.stnmPath = ''
        self.storage  = ''
        self.dtype    = ''
        self.title    = ''
        self.incre    = ''
        
        self.zrev     = False
        self.yrev     = False
        self.hasData  = False
        
        self.periodicX   = False
        self.cal365Days  = False
        self.template    = False
        self.sequential  = False
        
        self.totalZCount = 0
        self.zRecLength  = 0
        self.tRecLength  = 0
        self.byteOrder   = sys.byteorder
        
        if kwargs.get('file'):
            abspath = kwargs['file']
            
            if not '/' in abspath: # thanks to Baofeng Jiao from IAP
                abspath = './' + abspath
            
            if os.path.getsize(abspath) / (1024.0*1024.0) > 2:
                raise Exception('ctl file is too large (> 2 MB)')
            
            with open(abspath, 'r', encoding=encoding) as f:
                fileContent = f.readlines()
        
        elif kwargs.get('content'):
            abspath = None
            fileContent = kwargs['content'].splitlines()
        else:
            raise Exception('invalid key word '+
                            '(file='' or content='' is allowed)')
        
        for i, line in enumerate(fileContent):
            llower = line.lower()
            if (llower.startswith('dset') or
                llower.startswith('index') or
                llower.startswith('stnmap')
               ) and abspath is not None and '^' in line:
                dirname = os.path.dirname(abspath)
                if dirname[-1] != '/':
                    fileContent[i] = line.replace('^',
                                       os.path.dirname(abspath) + '/')
                else:
                    fileContent[i] = line.replace('^',
                                       os.path.dirname(abspath))
        
        self.descPath=abspath
        self.parse(fileContent)
    
    def parse(self, fileContent):
        """Parse file content as a multi-line str"""
        dpath_str = None
        
        for oneline in fileContent:
            if   oneline.lower().startswith('dset'):
                dpath_str = oneline.split()[1]
            elif oneline.lower().startswith('index'):
                self._processIndex(oneline)
            elif oneline.lower().startswith('stnmap'):
                self._processStnmap(oneline)
            elif oneline.lower().startswith('dtype'):
                self.dtype = oneline[5:].strip()
            elif oneline.lower().startswith('pdef'):
                self._processPDEF(oneline)
            elif oneline.lower().startswith('title'):
                self.title = oneline.split()[1].strip()
            elif oneline.lower().startswith('undef'):
                self.undef = float(oneline.split()[1].strip())
            elif oneline.lower().startswith('options'):
                self._processOptions(oneline)
            elif oneline.lower().startswith('byteswapped'):
                self.byteOrder  = 'big' \
                    if sys.byteorder == 'little' else 'little'
            elif oneline.lower().startswith('xdef'):
                self._processXDef(oneline, fileContent)
            elif oneline.lower().startswith('ydef'):
                self._processYDef(oneline,fileContent)
            elif oneline.lower().startswith('zdef'):
                self._processZDef(oneline, fileContent)
            elif oneline.lower().startswith('tdef'):
                self._processTDef(oneline)
            elif oneline.lower().startswith('edef'):
                self._processEDef(oneline, fileContent)
            elif oneline.lower().startswith('vars'):
                self._processVars(oneline, fileContent)
            elif oneline.lower().startswith('@ global string comment'):
                self._processGlobalComments(oneline)
            elif oneline.startswith('*') or oneline.strip() == '':
                continue
        
        if dpath_str == None:
            raise Exception('no valid dset is parsed')
        
        if self.template:
            self._processDSets(dpath_str)
        else:
            self._processDSet(dpath_str)
        
        if self.yrev:
            self.ydef.samples = np.flip(self.ydef.samples)
        
        if self.zrev:
            self.zdef = np.flip(self.zdef)
        
        if self.edef:
            strPos = 0
            for i, e in enumerate(self.edef):
                self.edef[i] = self.edef[i]._replace(strPos=strPos)
                strPos += e.tcount * self.tRecLength
    
    def _processDSets(self, dpath_str):
        strPos = dpath_str.find('%')
        
        if strPos == -1:
            raise Exception('template is used in ctl but no % in dset')
        
        endPos = len(dpath_str) - dpath_str[::-1].find('.') - 1
        # endPos = len(dpath_str) if endPos == -1 else endPos
        
        template = dpath_str[strPos:endPos]
        
        tokens = [self._get_template_format('%'+token)
                  for token in filter(lambda x: x != '', template.split('%'))]
        # tokens = []
        # for token in filter(lambda x: x != '', template.split('%')):
        #     tokens.append(self._get_template_format('%' + token))
        fmt = ''.join(tokens)
        
        fileList = []
        
        times = self.tdef.samples
        base  = self._get_field(times[0])
        for l in range(len(times)):
            part = times[l].item().strftime(fmt)
            
            fname = dpath_str[:strPos] + part + dpath_str[endPos:]
            fname = self._replace_forecast_template(fname, l, base)
            
            # remove duplicated file
            if fname not in fileList:
                fileList.append(fname)
        
        self.dsetPath = np.array(fileList)
        
        has = True
        for file in fileList:
            if not os.path.exists(file):
                has = False
                break

        self.hasData = has
        
    def _processDSet(self, dpath_str):
        self.dsetPath = dpath_str
        self.hasData  = os.path.exists(self.dsetPath)
    
    def _processIndex(self, oneline):
        self.indxPath = oneline.split()[1]
    
    def _processStnmap(self, oneline):
        self.stnmPath = oneline.split()[1]
    
    def _processPDEF(self, oneline):
        self.pdef = PDEF(oneline)
    
    def _processOptions(self, oneline):
        lineLower = oneline.lower()
        
        if 'yrev'             in lineLower: self.yrev       = True
        if 'zrev'             in lineLower: self.zrev       = True
        if 'template'         in lineLower: self.template   = True
        if 'sequential'       in lineLower: self.sequential = True
        if '365_day_calendar' in lineLower: self.cal365Days = True
        if 'big_endian'       in lineLower: self.byteOrder  = 'big'
        if 'byteswapped'      in lineLower: self.byteOrder  = \
            'big' if sys.byteorder == 'little' else 'little'
    
    def _processXDef(self, oneline, fileContent):
        tokens = oneline.lower().split()
        xnum   = int(tokens[1])
        
        if   tokens[2] == 'linear': xlnr = True
        elif tokens[2] == 'levels': xlnr = False
        else: raise Exception('invalid type for xdef (linear or levels): ' +
                              tokens[2])

        if xlnr:
            start, intv = float(tokens[3]), float(tokens[4])
            self.xdef = Coordinate('xdef', np.linspace(start,
                                                       start + intv * (xnum-1),
                                                       xnum, dtype=np.float32))
        else:
            values = [float(i) for i in tokens[3:]]
            count  = len(values)
            index  = fileContent.index(oneline) + 1
            
            while count < xnum:
                split = fileContent[index].split()
                values += [float(v) for v in split]
                count  += len(split)
                index  += 1
            
            if count != xnum:
                raise Exception('not parse xdef correctly')
            
            self.xdef = Coordinate('xdef', np.array(values))

        self.periodicX = self.xdef.isPeriodic(360)

    def _processYDef(self, oneline, fileContent):
        tokens = oneline.lower().split()
        ynum   = int(tokens[1])
        
        if   tokens[2] == 'linear': ylnr = True
        elif tokens[2] == 'levels': ylnr = False
        else: raise Exception('invalid type for ydef (linear or levels): ' +
                              tokens[2])
        
        if ylnr:
            start, intv = float(tokens[3]), float(tokens[4])
            self.ydef = Coordinate('ydef', np.linspace(start,
                                                       start + intv * (ynum-1),
                                                       ynum, dtype=np.float32))
        else:
            values = [float(i) for i in tokens[3:]]
            count  = len(values)
            index  = fileContent.index(oneline) + 1
            
            while count < ynum:
                split = fileContent[index].split()
                values += [float(v) for v in split]
                count  += len(split)
                index  += 1
            
            if count != ynum:
                raise Exception(('not parse ydef correctly, count={0} '+
                                 'while ynum={1}').format(count, ynum))
            
            self.ydef = Coordinate('ydef', np.array(values))
    
    def _processZDef(self, oneline, fileContent):
        tokens = oneline.lower().split()
        znum   = int(tokens[1])
        
        if   tokens[2] == 'linear': zlnr = True
        elif tokens[2] == 'levels': zlnr = False
        else: raise Exception('invalid type for zdef (linear or levels): ' +
                              tokens[2])
        
        if zlnr:
            start, intv = float(tokens[3]), float(tokens[4])
            self.zdef = Coordinate('zdef', np.linspace(start,
                                                       start + intv * (znum-1),
                                                       znum, dtype=np.float32))
            
        else:
            values = [float(i) for i in tokens[3:]]
            count  = len(values)
            index  = fileContent.index(oneline) + 1
            
            while count < znum:
                split = fileContent[index].split()
                values += [float(v) for v in split]
                count  += len(split)
                index  += 1
            
            if count != znum:
                raise Exception('zdef not parsed correctly')
            
            self.zdef = Coordinate('zdef', np.array(values))

    def _processTDef(self, oneline):
        tokens = oneline.lower().split()
        tnum   = int(tokens[1])

        if tokens[2]!='linear':
            raise Exception('nonlinear tdef is not supported')

        times = self._times_to_array(tokens[3], tokens[4], tnum)
        
        self.incre = GrADS_increment_to_timedelta64(tokens[4])
        self.tdef  = Coordinate('tdef', times)

    def _processEDef(self, oneline, fileContent):
        if not self.tdef:
            raise Exception('edef should be after tdef')
        
        from collections import namedtuple
        
        Ensemble = namedtuple('Ensemble', ['name', 'tcount', 'tstart',
                                           'codes', 'strPos'])
        
        tokens = oneline.lower().split()
        enum   = int(tokens[1])
        tdef   = self.tdef
        index  = fileContent.index(oneline) + 1
        
        if len(tokens) > 2:
            if tokens[2].lower() == 'names':
                NAMES = True
            else:
                raise Exception(f'Expected str \'NAMEs\', but found {tokens[2]}')
            
            if NAMES:
                enames = [i for i in tokens[3:]]
                count  = len(enames)
                
                while count < enum:
                    split = fileContent[index].split()
                    enames += [v for v in split]
                    count  += len(split)
                    index  += 1
                
                if count != enum:
                    raise Exception((f'edef not parsed correctly, count={count} '+
                                     'while enum={enum}'))
                
                self.edef = [Ensemble(name, tdef.length(), tdef.samples[0], None, 0)
                             for name in enames]
        else:
            tmp = []
            for i in range(enum):
                eline = fileContent[index + i].strip()
                
                if eline.lower() == 'endedef':
                    raise Exception(f'not enough ensembles, need {enum}')
                
                tkns  = eline.split()
                
                name = tkns[0]
                tcnt = int(tkns[1])
                tstr = GrADStime_to_datetime(tkns[2])
                code = None
                
                if len(tkns) >3:
                    code = tkns[3]
                    
                if tcnt > tdef.length():
                    raise Exception('Tcount in EDEF must be less than or equal' +
                                    'to the Tcount in TDEF')
                
                tmp.append(Ensemble(name, tcnt, tstr, code, 0))
            
            self.edef = tmp
    
    def _processVars(self, oneline, fileContent):
        if (self.dtype != 'station' and 
            not all([self.tdef, self.zdef, self.ydef, self.xdef])):
            raise Exception('vdef should be after x, y, z and t definitions')
        
        t = self.tdef.length()
        
        if self.dtype != 'station':
            y = self.ydef.length() if self.pdef is None else self.pdef.jsize
            x = self.xdef.length() if self.pdef is None else self.pdef.isize
        
            self.zRecLength = x * y * 4
            
            # add two bytes at the beginning and the end
            if self.sequential:
                self.zRecLength += 8
        
        tokens = oneline.split()
        vnum   = int(tokens[1])
        start  = fileContent.index(oneline) + 1
        
        if vnum < 1:
            raise Exception('there should be at least one CtlVar')
        
        self.vcount = vnum
        self.vdef   = [None] * vnum
        
        v = CtlVar(fileContent[start])
        v.index =0
        v.strPos=0
        v.tcount=t
        
        if self.dtype != 'station':
            v.ycount=y
            v.xcount=x
        
        self.totalZCount += v.zcount
        self.storage      = v.storage
        self.vdef[0]      = v
        
        type1 = type2 = type3 = False
        
        for i in range(1, vnum):
            v  = CtlVar(fileContent[start+i])
            vs = self.vdef[i-1]
            
            v.index  = i
            v.undef  = self.undef
            v.tcount = t
            
            if self.dtype != 'station':
                v.ycount = y
                v.xcount = x
            
            # if v.storage in ['99', '0']:
            if v.storage in ['99', '0', '00', '000', '1', '11', '111']:
                v.strPos = vs.zcount * self.zRecLength + vs.strPos
                type1 = True
            elif v.storage == '-1,20':
                v.strPos = vs.zcount * self.zRecLength * t + vs.strPos
                type2 = True
            else:
                type3 = True
            
            if not type3 and type1 == type2:
                raise Exception('storage type should be the same')
            
            self.totalZCount += v.zcount
            self.vdef[i] = v
        
        self.tRecLength = self.zRecLength * self.totalZCount
        
        if fileContent[start + vnum].strip().lower() != 'endvars':
            raise Exception('endvars is expected')
    
    def _processGlobalComments(self, oneline):
        cnt = oneline[24:].strip().split('=')
        
        self.comments[cnt[0].strip()] = cnt[1].strip()

    def _get_template_format(self, part):
        """Get time format string
        
        See the following URL for reference:
        http://cola.gmu.edu/grads/gadoc/templates.html
        
        %x1   1 digit decade
        %x3   3 digit decade
        %y2   2 digit year
        %y4   4 digit year
        %m1   1 or 2 digit month
        %m2   2 digit month (leading zero if needed)
        %mc   3 character month abbreviation
        %d1   1 or 2 digit day
        %d2   2 digit day (leading zero if needed)
        %h1   1 or 2 digit hour
        %h2   2 digit hour
        %h3   3 digit hour (e.g., 120 or 012)
        %n2   2 digit minute; leading zero if needed
        %f2   2 digit forecast hour; leading zero if needed; more digits added
              for hours >99; hour values increase indefinitely
        %f3   3 digit forecast hour; leading zeros if needed; more digits added
              for hours >999; hour values increase indefinitely
        %fn2  2 digit forecast minute; leading zero if needed; more digits added
              for minutes > 99; minute values increase indefinitely (2.0.a9+)
        %fhn  forecast time expressed in hours and minutes (hhnn) where minute
              value (nn) is always <=59 and hour value (hh) increases indefinitely.
              If hh or nn are <=9, they are padded with a 0, so they are always at
              least 2 digits; more digits added for hours >99. (2.0.a9+)
        %fdhn forecast time expressed in days, hours, and minutes (ddhhnn) where
              minute value (nn) is always <=59, hour value (hh) is always <=23 and
              day value (dd) increases indefinitely. If dd, hh, or nn are <=9, they
              are padded with a 0 so they are always at least 2 digits; more digits
              added for days >99. (2.0.a9+)
        %j3   3 digit julian day (day of year) (2.0.a7+)
        %t1   1 or 2 digit time index (file names contain number sequences that
              begin with 1 or 01) (2.0.a7+)
        %t2   2 digit time index (file names contain number sequences that begin
              with 01) (2.0.a7+)
        %t3   3 digit time index (file names contain number sequences that begin
              with 001) (2.0.a7+)
        %t4   4 digit time index (file names contain number sequences that begin
              with 0001) (2.0.a8+)
        %t5   5 digit time index (file names contain number sequences that begin
              with 00001) (2.0.a8+)
        %t6   6 digit time index (file names contain number sequences that begin
              with 000001) (2.0.a8+)
        %tm1  1 or 2 digit time index (file names contain number sequences that
              begin with 0 or 00) (2.0.a7+)
        %tm2  2 digit time index (file names contain number sequences that begin
              with 00) (2.0.a7+)
        %tm3  3 digit time index (file names contain number sequences that begin
              with 000) (2.0.a7+)
        %tm4  4 digit time index (file names contain number sequences that begin
              with 0000) (2.0.a8+)
        %tm5  5 digit time index (file names contain number sequences that begin
              with 00000) (2.0.a8+)
        %tm6  6 digit time index (file names contain number sequences that begin
              with 000000) (2.0.a8+)
    
        Parameters
        ----------
        part : str
            A string in the above format started with %.
    
        Returns
        -------
        str
            The format in python datetime
        """
        if   part == '%y2':
            return '%y'
        elif part == '%y4':
            return '%Y'
        elif part == '%m1':
            return '%m'
        elif part == '%m2':
            return '%m'
        elif part == '%mc':
            return '%b'
        elif part == '%d1':
            return '%d'
        elif part == '%d2':
            return '%d'
        elif part == '%h1':
            return '%H'
        elif part == '%h2':
            return '%H'
        elif part == '%n2':
            return '%M'
        elif part in ['%f3', '%f2', '%fn2', '%fhn', '%fdhn']:
            # this is not supported by strftime()
            return '_miniufo_' + part[1:]
        else:
            raise Exception('unsupported format: ' + part)
    
    def _replace_forecast_template(self, fname, l, base):
        """Replace forecast str %f as a template in dset
    
        Parameters
        ----------
        fname: str
            A given string of binary file.
        l: int
            Index of file in a template.
        base: numpy.datetime64
            Base time.
    
        Returns
        -------
        str
            A string after replacing the %f template.
        """
        amount = l * self.incre
        
        days = base[..., 2]
        hour = base[..., 3]
        mins = base[..., 4]
        
        if fname.find('_miniufo_f3') != -1:
            dt = (amount.astype("timedelta64[h]") + hour).astype(int)
            fname = fname.replace('_miniufo_f3', '{0:03d}'.format(dt))
        
        if fname.find('_miniufo_f2') != -1:
            dt = (amount.astype("timedelta64[h]") + hour).astype(int)
            fname = fname.replace('_miniufo_f2', '{0:02d}'.format(dt))
        
        if fname.find('_miniufo_fn2') != -1:
            dt = (amount.astype("timedelta64[m]") + mins).astype(int)
            fname = fname.replace('_miniufo_fn2', '{0:02d}'.format(dt))
        
        if fname.find('_miniufo_fhn') != -1:
            dt_h = (amount.astype("timedelta64[h]") + hour).astype(int)
            dt_m = (amount.astype("timedelta64[m]") + mins).astype(int) % 60
            fname = fname.replace('_miniufo_fhn',
                                  '{0:02d}'.format(dt_h)+'{0:02d}'.format(dt_m))
        
        if fname.find('_miniufo_fdhn') != -1:
            dt_d = (amount.astype("timedelta64[D]") + days).astype(int)
            dt_h = (amount.astype("timedelta64[h]") + hour).astype(int) % 24
            dt_m = (amount.astype("timedelta64[m]") + mins).astype(int) % 60
            fname = fname.replace('_miniufo_fdhn',
                                  '{0:02d}'.format(dt_d)+
                                  '{0:02d}'.format(dt_h)+
                                  '{0:02d}'.format(dt_m))
        
        return fname

    def _get_field(self, datetime64):
        """Get fields of a datetime64
        
        Convert array of datetime64 to a calendar array of
        [year, month, day, hour, minute, seconds, microsecond]
        with these quantites indexed on the last axis.
    
        Parameters
        ----------
        datetime64: datetime64 array (...)
            numpy.ndarray of datetimes of arbitrary shape
    
        Returns
        -------
        cal: uint32 array (..., 7)
            calendar array with last axis representing year, month, day, hour,
            minute, second, microsecond
        """
        # allocate output 
        out = np.empty(datetime64.shape + (7,), dtype="u4")
        
        # decompose calendar floors
        Y, M, D, h, m, s = [datetime64.astype(f'M8[{x}]') for x in 'YMDhms']
        
        out[..., 0] = Y + 1970 # Gregorian Year
        out[..., 1] = (M - Y) + 1 # month
        out[..., 2] = (D - M) + 1 # dat
        out[..., 3] = (datetime64 - D).astype('m8[h]' ) # hour
        out[..., 4] = (datetime64 - h).astype('m8[m]' ) # minute
        out[..., 5] = (datetime64 - m).astype('m8[s]' ) # second
        out[..., 6] = (datetime64 - s).astype('m8[us]') # microsecond
        
        return out


    def _times_to_array(self, strTime, incre, tnum):
        """Change format of time
        
        Convert GrADS time string of strart time and increment
        to an array of numpy.datetime64.
        
        Parameters
        ----------
        strTime : str
            Grads start time e.g., 00:00z01Jan2000.
        incre : str
            Grads time increment in str format e.g., 1dy.
        tnum : int
            Grads time increment in str format e.g., 1dy.
        
        Returns
        -------
        numpy array of datetime64
            Times in datetime64 format
        """
        if 'mo' in incre:
            start = GrADStime_to_datetime(strTime)
            
            lst = []
            for l in range(tnum):
                y, m = start.year, start.month
                y, m = y+int((m+l-1)/12), int((m+l-1)%12)+1
                lst.append(start.replace(year=y, month=m))
            
            return np.asarray(lst, dtype='datetime64[s]')
            
        elif 'yr' in incre:
            start = GrADStime_to_datetime(strTime)
            
            lst = []
            for l in range(tnum):
                y = start.year + l
                lst.append(start.replace(year=y))
            
            return np.asarray(lst, dtype='datetime64[s]')
        
        else:
            start = GrADStime_to_datetime64(strTime)
            intv  = GrADS_increment_to_timedelta64(incre)

            if self.cal365Days:

                lst = []
                while len(lst) < tnum:
                    isfeb29 = (start.item().day == 29) and (start.item().month == 2)
                    if not isfeb29:
                        lst.append(start)
                    start += intv

                return np.asarray(lst)

            else:

                return np.arange(start, start + intv * tnum, intv)

    def __repr__(self):
        """Print this class as a string"""
        vdef = np.array(self.vdef)
        pdef = self.pdef.proj if self.pdef is not None else ''
        
        if self.edef is not None:
            edef = [f'Ensem: {e.name}, {e.tcount}, {e.tstart}, {e.codes}, {e.strPos}'
                    for e in self.edef]
        else:
            edef = 'None'
        
        return \
            '   dsetPath: ' + str(self.dsetPath)  + '\n'\
            '   descPath: ' + str(self.descPath)  + '\n'\
            '   indxPath: ' + str(self.indxPath)  + '\n'\
            '   stnmPath: ' + str(self.stnmPath)  + '\n'\
            '      title: ' + str(self.title)     + '\n'\
            '      undef: ' + str(self.undef)     + '\n'\
            '       zrev: ' + str(self.zrev)      + '\n'\
            '       yrev: ' + str(self.yrev)      + '\n'\
            '      dtype: ' + str(self.dtype)     + '\n'\
            '   template: ' + str(self.template)  + '\n'\
            '  periodicX: ' + str(self.periodicX) + '\n'\
            ' cal365Days: ' + str(self.cal365Days)+ '\n'\
            ' sequential: ' + str(self.sequential)+ '\n'\
            '  byteOrder: ' + str(self.byteOrder) + '\n'\
            '       xdef: ' + str(self.xdef)      + '\n'\
            '       ydef: ' + str(self.ydef)      + '\n'\
            '       zdef: ' + str(self.zdef)      + '\n'\
            '       tdef: ' + str(self.tdef)      + '\n'\
            '       edef: ' + str(edef)           + '\n'\
            '       pdef: ' + str(pdef)           + '\n'\
            '       vdef: ' + str(vdef)



class PDEF(object):
    """PDEF class
    
    Parse necessary info in PDEF.
    
    Reference: http://cola.gmu.edu/grads/gadoc/pdef.html
    
    Attributes
    ----------
    isize: int
        size of native grid in x direction
    jsize: int
        size of native grid in y direction
    proj: str
        type of projection
    lonref: str
        reference longitude
    """
    def __init__(self, oneline):
        """Constructor
        
        Parameters
        ----------
        oneline: str
            The ASCII line of PDEF in ctl file.
        """
        lineLower = oneline.lower()
        
        if 'nps' in lineLower or 'sps' in lineLower:
            token = lineLower.split()
            
            if len(token) != 8:
                raise Exception('not enough tokens for PDEF, ' +
                                'expected 8 but found ' + str(len(token)))
            
            self.isize   = int  (token[1]) # size of native grid in x direction
            self.jsize   = int  (token[2]) # size of native grid in y direction
            self.proj    =      (token[3]) # type of projection
            self.ipole   = int  (token[4]) # i-coord of pole ref to ll corner
            self.jpole   = int  (token[5]) # j-coord of pole ref to ll corner
            self.lonref  = float(token[6]) # reference longitude
            self.gridinc = float(token[7]) # distance between gripoints in km
            
        elif 'lccr' in lineLower or 'lcc' in lineLower:
            token = lineLower.split()
            
            if len(token) != 13:
                raise Exception('not enough tokens for PDEF, ' +
                                'expected 13 but found ' + str(len(token)))
            
            self.isize   = int  (token[1]) # size of native grid in x direction
            self.jsize   = int  (token[2]) # size of native grid in y direction
            self.proj    =      (token[3]) # type of projection
            self.latref  = float(token[4]) # ref latitude
            self.lonref  = float(token[5]) # ref longitude (E>0, W<0)
            self.iref    = float(token[6]) # i of ref point
            self.jref    = float(token[7]) # j of ref point
            self.Struelat= float(token[8]) # S true lat
            self.Ntruelat= float(token[9]) # N true lat
            self.slon    = float(token[10]) # standard longitude
            self.dx      = float(token[11]) # grid X increment in meters
            self.dy      = float(token[12]) # grid Y increment in meters
        
        else:
            raise Exception('not currently supported PDEF\n' + oneline)

    def __repr__(self):
        """Print this class as a string"""
        return '\n'.join(['%s: %s' % item for item in self.__dict__.items()])
    


class Coordinate(object):
    """Discrete sampled coordinate
    
    This is a simple wrapper for np.array for a coordinate.
    
    Attributes
    ----------
    isLinear: bool
        Steps are even or uneven
    isIncre: int
        Is increasing or decreasing
    name: str
        The name of the coordinate
    samples: str
        Discretized coordinate samples
    delSamples: str
        Finite difference between samples
    """
    def __init__(self, name, samples):
        """Constructor
        
        Parameters
        ----------
        name : str
            The name of the coordinate.
        samples : np.array
            1D array for the discrete coordinate.
        """
        self.isLinear   = True
        self.isIncre    = True
        self.name       = name
        self.samples    = samples
        self.delSamples = None
        
        self.max = np.max(self.samples)
        self.min = np.min(self.samples)
        
        if len(samples) > 1:
            self.delSamples = np.diff(self.samples)
            
            if self.samples[-1] < self.samples[0]:
                self.isIncre=False
        else:
            self.delSamples = np.array([1])
    
    def length(self):
        return len(self.samples)
    
    def isPeriodic(self,period):
        # not physically but generally true
        if not self.isLinear: return False
        
        delta = self.delSamples[0]
        start = self.samples[-1] + delta - period
        
        if(abs((start - self.samples[0]) / delta > 1e-4)):
            return False
        
        return True
    
    def __str__(self):
        """Print this class as a string"""
        return str(self.samples)



class CtlVar(object):
    """A simple variable class used in .ctl file
    
    Attributes
    ----------
    tcount: int
        T grid points
    zcount: int
        Z grid points
    ycount: int
        Y grid points
    xcount: int
        X grid points
    undef: float
        Undefined values
    dependZ: bool
        Whether the var depends on z
    unit: str
        Unit of the variable
    name: str
        Name of the variable
    comment: str
        A short comment
    index: str
        Index of this variable
    strPos: str
        Start position (in bytes) of this variable in the binary file
    """
    __reBlank = re.compile(r'[\s\t]+')
    __reUnits = re.compile(r'\([^\(\)]+?\)')
    
    
    def __init__(self, oneLineStr):
        self.tcount = 0
        self.zcount = 0
        self.ycount = 0
        self.xcount = 0
        self.undef  = np.nan
        self.dependZ= True # whether the var depends on z
        
        self.unit   = ''
        self.name   = ''
        self.comment= ''
        
        self.index  = 0
        self.strPos = 0
        
        if len(CtlVar.__reBlank.split(oneLineStr.strip(), maxsplit=3))== 3:
            self.name, self.zcount, self.storage = \
                CtlVar.__reBlank.split(oneLineStr.strip(), maxsplit=3)
            self.comment = self.name
        else:
            self.name, self.zcount, self.storage, self.comment = \
                CtlVar.__reBlank.split(oneLineStr.strip(), maxsplit=3)
        
        self.zcount = int(self.zcount)
        
        findMatch = CtlVar.__reUnits.findall(self.comment)
        
        if findMatch:
            self.unit    = findMatch[-1]
            self.comment = self.comment[:self.comment.index(self.unit)].strip()
        else:
            self.unit = ''

        if self.zcount == 0:
            self.zcount = 1
            self.dependZ= False
    
    def __str__(self):
        """
        Print this class as a string.
        """
        return '\n'.join(('%8s: %s' % item for item in self.__dict__.items()))
    
    def __repr__(self):
        """
        Print this class as a string.
        """
        return 'CtlVar: {0:8s} in shape (t={1:d}, z={2:d}, y={3:d}, x={4:d})'\
                .format(self.name,
                        self.tcount, self.zcount,
                        self.ycount, self.xcount)


"""
Some useful functions defined here
"""
def GrADStime_to_datetime(gradsTime):
    """Convert GrADS time string e.g., 00:00z01Jan2000 to datetime
    
    Parameters
    ----------
    gradsTime: str
        Grads time in str format e.g., 00:00z01Jan2000.
    
    Returns
    -------
    datetime
        GrADS time in datetime format
    """
    lens = len(gradsTime)
    
    if   lens==15 or lens==14:
        time = datetime.strptime(gradsTime, "%H:%Mz%d%b%Y")
    elif lens==12 or lens==11:
        time = datetime.strptime(gradsTime, "%Hz%d%b%Y"   )
    elif lens== 9 or lens== 8:
        time = datetime.strptime(gradsTime, "%d%b%Y"      )
    elif lens== 7:
        time = datetime.strptime(gradsTime, "%b%Y"        )
    else:
        raise Exception('invalid length of GrADS date/time string')
    
    return time


def GrADStime_to_datetime64(gradsTime):
    """Convert GrADS time string e.g., 00:00z01Jan2000 to numpy.datetime64
    
    Parameters
    ----------
    gradsTime : str
        Grads time in str format e.g., 00:00z01Jan2000.
    
    Returns
    -------
    datetime64
        GrADS time in datetime64 format
    """
    time = GrADStime_to_datetime(gradsTime)
    
    return datetime64(time.strftime('%Y-%m-%dT%H:%M:%S'))


def GrADS_increment_to_timedelta64(incre):
    """Convert GrADS time increment string to numpy.timedelta64
    
    Parameters
    ----------
    incre: str
        Grads time increment in str format e.g., 1dy.
    
    Returns
    -------
    timedelta64
        GrADS time in datetime64 format
    """
    unit   = incre[-2:]
    amount = incre[:-2]

    unitDict = {
        'se': 's',
        'mn': 'm',
        'hr': 'h',
        'dy': 'D',
        'mo': 'M',
        'yr': 'Y'}

    return timedelta64(int(amount), unitDict[unit])







# -*- coding: utf-8 -*-
"""
Created on 2020.04.11

@author: MiniUFO
Copyright 2018. All rights reserved. Use is subject to license terms.
"""
import os
import numpy as np
import xarray as xr
import dask.array as dsa
from dask.base import tokenize
from glob import glob
from pathlib import Path
from .core import CtlDescriptor
from functools import reduce

"""
IO related functions here
"""
def open_mfdataset(paths, parallel=False, encoding='GBK'):
    """Open multiple ctl files as a single dataset
    
    This is similar to `xarray.open_mfdataset()` that can simutaneously open
    a series of ctl files that are in similar spatial ranges but of different
    temporal ranges.

    Parameters
    ----------
    paths: str or sequence
        Either a string glob in the form ``"path/to/my/files/*.ctl"`` or an
        explicit list of files to open. Paths can be given as strings or as
        pathlib Paths.
    parallel: bool, optional
        If True, the open and preprocess steps of this function will be
        performed in parallel using ``dask.delayed``. Default is False.
    encoding: str
        Encoding for the ctl file content e.g., ['GBK', 'UTF-8'].

    Returns
    -------
    re: xarray.Dataset
        A dataset.

    Notes
    -----
    ``open_mfdataset`` opens files with read-only access. When you modify values
    of a Dataset, even one linked to files on disk, only the in-memory copy you
    are manipulating in xarray is modified: the original file on disk is never
    touched.
    """
    if isinstance(paths, str):
        paths = sorted(glob(paths))
    else:
        paths = [str(p) if isinstance(p, Path) else p for p in paths]

    if not paths:
        raise OSError("no files to open")

    if parallel:
        import dask

        # wrap the open_dataset
        open_ = dask.delayed(open_CtlDataset)
    else:
        open_ = open_CtlDataset

    datasets = [open_(p, encoding=encoding) for p in paths]

    if parallel:
        # calling compute here will return the datasets/file_objs lists,
        # the underlying datasets will still be stored as dask arrays
        datasets = dask.compute(datasets)

        return xr.concat(datasets[0], dim='time')

    combined = xr.concat(datasets, dim='time')

    return combined


def open_CtlDataset(desfile, returnctl=False, encoding='GBK'):
    """Open a single ctl dataset
    
    Open a 4D dataset with a descriptor file end with .ctl and
    return a xarray.Dataset.  This also uses the dask to chunk
    very large dataset, which is similar to the xarray.open_dataset.

    Parameters
    ----------
    desfile: string
        Path to the descriptor file end with .ctl or .cts
    returnctl: bool
        Return dset and ctl as a tuple

    Returns
    -------
    dset: xarray.Dataset
        Dataset object containing all coordinates and variables.
    ctl: xgrads.CtlDescriptor
        Ctl descriptor file if returnctl == True.
    """
    if isinstance(desfile, str):
        if not desfile.endswith('.ctl'):
            raise Exception('unsupported file, suffix should be .ctl')

        ctl = CtlDescriptor(encoding=encoding, file=desfile)
    elif isinstance(desfile, CtlDescriptor):
        ctl = desfile
    else:
        raise Exception('unsupported type of input ('+str(type(desfile))+'), ' +
                        '[CtlDescriptor or str] are allowed')

    if ctl.template:
        if ctl.edef:
            raise Exception('template and EDEF are not simultaneously supported yet')
        
        tcount = len(ctl.tdef.samples) # number of total time count
        tcPerf = []                    # number of time count per file
        
        has_missing = False
        for file in ctl.dsetPath:
            if os.path.exists(file):
                fsize = os.path.getsize(file)
    
                if fsize % ctl.tRecLength != 0:
                    raise Exception('incomplete file for ' + file +
                                    ' (not multiple of ' + str(ctl.tRecLength) +
                                    ' bytes)')
    
                tcPerf.append(fsize // ctl.tRecLength)
            else:
                print(' warning: ' + file + ' is missing...')
                has_missing = True
        
        if has_missing:
            raise Exception('there are missing binary files')

        total_size = sum(tcPerf)

        if total_size < tcount:
            raise Exception('no enough files for ' + str(tcount) +
                            ' time records')

        # get time record number in each file
        rem = tcount
        idx = 0
        for i, num in enumerate(tcPerf):
            rem -= num
            if rem <= 0:
                idx = i
                break

        tcPerf_m      = tcPerf[:idx+1]
        tcPerf_m[idx] = tcPerf[idx   ] + rem

        # print(ctl.dsetPath)
        # print(tcPerf)
        # print(tcPerf_m)

        binData = __read_template_as_dask(ctl, tcPerf_m)

    else:
        if ctl.edef == None:
            expect = ctl.tRecLength * ctl.tdef.length()
        else:
            expect = 0
            for ens in ctl.edef:
                expect += ens.tcount * ctl.tRecLength
        
        actual = os.path.getsize(ctl.dsetPath)

        if expect != actual:
            print('WARNING: expected binary file size: {0}, actual size: {1}'
                            .format(expect, actual))

        binData = __read_as_dask(ctl)

    variables = []

    if ctl.pdef is None:
        for m, v in enumerate(ctl.vdef):
            if v.dependZ:
                if ctl.edef:
                    data = binData[m]
                    dims = ['ens', 'time', 'lev', 'lat', 'lon']
                    coords = {'ens' : [ens.name for ens in ctl.edef],
                              'time': ctl.tdef.samples[:],
                              'lev' : ctl.zdef.samples[:v.zcount],
                              'lat' : ctl.ydef.samples[:],
                              'lon' : ctl.xdef.samples[:]}
                else:
                    data = binData[m]
                    dims = ['time', 'lev', 'lat', 'lon']
                    coords = {'time': ctl.tdef.samples[:],
                              'lev' : ctl.zdef.samples[:v.zcount],
                              'lat' : ctl.ydef.samples[:],
                              'lon' : ctl.xdef.samples[:]}
            else:
                if ctl.edef:
                    e, t, z, y, x = binData[m].shape
                    data = binData[m].reshape((e,t,y,x))
                    dims = ['ens', 'time', 'lat', 'lon']
                    coords = {'ens' : [ens.name for ens in ctl.edef],
                              'time': ctl.tdef.samples[:],
                              'lat' : ctl.ydef.samples[:],
                              'lon' : ctl.xdef.samples[:]}
                else:
                    t, z, y, x = binData[m].shape
                    data = binData[m].reshape((t,y,x))
                    dims = ['time', 'lat', 'lon']
                    coords = {'time': ctl.tdef.samples[:],
                              'lat' : ctl.ydef.samples[:],
                              'lon' : ctl.xdef.samples[:]}
            
            da = xr.DataArray(name=v.name, data=data, dims=dims, coords=coords,
                              attrs={'comment': v.comment,
                                     'storage': v.storage})
            variables.append(da)

    else:
        PDEF = ctl.pdef

        if PDEF.proj in ['lcc', 'lccr']:
            ycoord = np.linspace(0, (PDEF.jsize-1) * PDEF.dx, PDEF.jsize)
            xcoord = np.linspace(0, (PDEF.isize-1) * PDEF.dy, PDEF.isize)

        elif PDEF.proj in ['sps', 'nps']:
            inc = PDEF.gridinc * 1000 # change unit from km to m
            ycoord = np.linspace(-(PDEF.jpole), (PDEF.jsize-PDEF.jpole),
                                   PDEF.jsize) * inc
            xcoord = np.linspace(-(PDEF.ipole), (PDEF.isize-PDEF.ipole),
                                   PDEF.isize) * inc

        for m, v in enumerate(ctl.vdef):
            if v.dependZ:
                if ctl.edef:
                    data = binData[m]
                    dims = ['ens', 'time', 'lev', 'y', 'x']
                    coords = {'ens' : [ens.name for ens in ctl.edef],
                              'time': ctl.tdef.samples[:],
                              'lev' : ctl.zdef.samples[:v.zcount],
                              'y'   : ycoord,
                              'x'   : xcoord}
                else:
                    data = binData[m]
                    dims = ['time', 'lev', 'y', 'x']
                    coords = {'time': ctl.tdef.samples[:],
                              'lev' : ctl.zdef.samples[:v.zcount],
                              'y' : ycoord,
                              'x' : xcoord}
            else:
                if ctl.edef:
                    e, t, z, y, x = binData[m].shape
                    data = binData[m].reshape((e,t,y,x))
                    dims = ['ens', 'time', 'y', 'x']
                    coords = {'ens' : [ens.name for ens in ctl.edef],
                              'time': ctl.tdef.samples[:],
                              'y' : ycoord,
                              'x' : xcoord}
                else:
                    t, z, y, x = binData[m].shape
                    data = binData[m].reshape((t,y,x))
                    dims = ['time', 'y', 'x']
                    coords = {'time': ctl.tdef.samples[:],
                              'y' : ycoord,
                              'x' : xcoord}

            da = xr.DataArray(name=v.name, data=data, dims=dims, coords=coords,
                              attrs={'comment': v.comment,
                                     'storage': v.storage})
            variables.append(da)

#    variables = {v.name: (['time','lev','lat','lon'], binData[m])
#                 for m,v in enumerate(ctl.vdef)}

    dset = xr.merge(variables)

    dset.attrs['title'] = ctl.title
    dset.attrs['undef'] = ctl.undef
    dset.attrs['pdef' ] = 'None'

    if ctl.pdef:
        dset.attrs['pdef' ] = ctl.pdef.proj

    if returnctl:
        return dset, ctl
    else:
        return dset



"""
Helper (private) methods are defined below
"""
def __read_as_dask(dd):
    """Read binary data and return as a dask array

    Parameters
    ----------
    dd: CtlDescriptor
        A CtlDescriptor

    Returns
    -------
    re: dask.array
        Data viewed as `dask.array`.
    """
    if dd.pdef is None:
        t, y, x = dd.tdef.length(), dd.ydef.length(), dd.xdef.length()
    else:
        t, y, x = dd.tdef.length(), dd.pdef.jsize, dd.pdef.isize

    totalNum = sum([reduce(lambda x, y:
                    x*y, (t,v.zcount,y,x)) for v in dd.vdef])

    if dd.sequential:
        sequentialSize = x * y + 2
    else:
        sequentialSize = -1

    # print(totalNum * 4.0 / 1024.0 / 1024.0)

    binData = []

    dtype   = '<f4' if dd.byteOrder == 'little' else '>f4'

    for m, v in enumerate(dd.vdef):
        name = '@miniufo_' + tokenize(v, m)

        if totalNum < (100 * 100 * 100 * 10): # about 40 MB, chunk all
            # print('small')
            if dd.edef == None:
                chunk = (t, v.zcount, y, x)
                shape = (t, v.zcount, y, x)
    
                dsk = {(name, 0, 0, 0, 0):
                       (__read_var, dd.dsetPath, v, 0, dd.tRecLength,
                        None, None, dtype, sequentialSize)}
            else:
                e = len(dd.edef)
                chunk = (1, t, v.zcount, y, x)
                shape = (e, t, v.zcount, y, x)
    
                dsk = {(name, n, 0, 0, 0, 0):
                       (__read_var, dd.dsetPath, v, ens.strPos, dd.tRecLength,
                        None, None, dtype, sequentialSize)
                       for n, ens in enumerate(dd.edef)}

            binData.append(dsa.Array(dsk, name, chunk,
                                     dtype=dtype, shape=shape))

        elif totalNum > (200 * 100 * 100 * 100): # about 800 MB, chunk 2D slice
            # print('large')
            if dd.edef == None:
                chunk = (1, 1, y, x)
                shape = (t, v.zcount, y, x)

                dsk = {(name, l, k, 0, 0):
                       (__read_var, dd.dsetPath, v, 0, dd.tRecLength,
                        l, k, dtype, sequentialSize)
                       for l in range(t)
                       for k in range(v.zcount)}
            else:
                e = len(dd.edef)
                chunk = (1, 1, 1, y, x)
                shape = (e, t, v.zcount, y, x)

                dsk = {(name, n, l, k, 0, 0):
                       (__read_var, dd.dsetPath, v, ens.strPos, dd.tRecLength,
                        l, k, dtype, sequentialSize)
                       for n, ens in enumerate(dd.edef)
                       for l in range(t)
                       for k in range(v.zcount)}

            binData.append(dsa.Array(dsk, name, chunk,
                                     dtype=dtype, shape=shape))


        else: # in between, chunk 3D slice
            # print('between')
            if dd.edef == None:
                chunk = (1, v.zcount, y, x)
                shape = (t, v.zcount, y, x)
    
                dsk = {(name, l, 0, 0, 0):
                       (__read_var, dd.dsetPath, v, 0, dd.tRecLength,
                        l, None, dtype, sequentialSize)
                       for l in range(t)}
            else:
                e = len(dd.edef)
                chunk = (1, 1, v.zcount, y, x)
                shape = (e, t, v.zcount, y, x)
    
                dsk = {(name, n, l, 0, 0, 0):
                       (__read_var, dd.dsetPath, v, ens.strPos, dd.tRecLength,
                        l, None, dtype, sequentialSize)
                       for n, ens in enumerate(dd.edef)
                       for l in range(t)}

            binData.append(dsa.Array(dsk, name, chunk,
                                     dtype=dtype, shape=shape))

    return binData


def __read_template_as_dask(dd, tcPerf):
    """Read template binary data and return as a dask array

    Parameters
    ----------
    dd: CtlDescriptor
        A CtlDescriptor
    tcPerf: list of int
        Number of time count per file

    Returns
    -------
    re: list of dask.array
        Data viewed as `dask.array`.
    """
    if dd.pdef is None:
        t, y, x = dd.tdef.length(), dd.ydef.length(), dd.xdef.length()
    else:
        t, y, x = dd.tdef.length(), dd.pdef.jsize, dd.pdef.isize

    totalNum = sum([reduce(lambda x, y:
                    x*y, (tcPerf[0],v.zcount,y,x)) for v in dd.vdef])

    if dd.sequential:
        sequentialSize = x * y + 2
    else:
        sequentialSize = -1

    # print(totalNum * 4.0 / 1024.0 / 1024.0)

    binData = []

    dtype   = '<f4' if dd.byteOrder == 'little' else '>f4'

    for m, v in enumerate(dd.vdef):
        name = '@miniufo_' + tokenize(v, m)

        if totalNum > (200 * 100 * 100 * 100): # about 800 MB, chunk 2D slice
            # print('large')
            chunk = (1, 1, y, x)
            shape = (t, v.zcount, y, x)

            dsk = {(name, l + sum(tcPerf[:m]), k, 0, 0):
                   (__read_var, f, v, 0, dd.tRecLength,
                    l, k, dtype, sequentialSize)
                   for m, f in enumerate(dd.dsetPath[:len(tcPerf)])
                   for l in range(tcPerf[m])
                   for k in range(v.zcount)}

            binData.append(dsa.Array(dsk, name, chunk,
                                     dtype=dtype, shape=shape))

        else: # in between, chunk 3D slice
            # print('between')
            chunk = (1, v.zcount, y, x)
            shape = (t, v.zcount, y, x)

            dsk = {(name, l + sum(tcPerf[:m]), 0, 0, 0):
                   (__read_var, f, v, 0, dd.tRecLength,
                    l, None, dtype, sequentialSize)
                   for m, f in enumerate(dd.dsetPath[:len(tcPerf)])
                   for l in range(tcPerf[m])}

            binData.append(dsa.Array(dsk, name, chunk,
                                     dtype=dtype, shape=shape))

    return binData


def __read_var(file, var, epos, tstride, tstep, zstep, dtype, sequentialSize=-1):
    """Read a variable given the trange

    Parameters
    ----------
    file: str
        A file from which data are read.
    var: CtlVar
        A variable that need to be read.
    epos: int
        Position of the ensemble dimension
    tstride: int
        Stride of a single time record.
    tstep: int
        T-step to be read, started from 0.  If None, read all t-steps
    zstep: int
        Z-step to be read, started from 0.  If None, read all z-steps
    sequentialSize: int
        Size of the sequential block (= y * x).  Default of -1 means
        non-sequential storage.

    Returns
    -------
    re: numpy.ndarray
        Binary data.
    """
    # print(var.name+' '+str(tstep)+' '+str(zstep)+' '+str(var.strPos))

    if var.storage == '-1,20':
        if tstep is None and zstep is None:
            shape = (var.tcount, var.zcount, var.ycount, var.xcount)
            if sequentialSize != -1:
                seqShp = (var.tcount, var.zcount, sequentialSize)
            else:
                seqShp = shape
            pos   = var.strPos + epos
            return __read_continuous(file, pos, shape, dtype,
                                     sequentialShape=seqShp)

        elif zstep is None and tstep is not None:
            shape = (1, var.zcount, var.ycount, var.xcount)
            if sequentialSize != -1:
                seqShp = (1, var.zcount, sequentialSize)
            else:
                seqShp = shape
            pos   = var.strPos + epos
            return __read_continuous(file, pos, shape, dtype,
                                     sequentialShape=seqShp)

        elif tstep is None and zstep is not None:
            raise Exception('not implemented in -1,20')

        else:
            shape = (1, 1, var.ycount, var.xcount)
            zstri = var.ycount * var.xcount * 4
            if sequentialSize != -1:
                seqShp = (1, 1, sequentialSize)
                zstri += 8
            else:
                seqShp = shape
            pos   = var.strPos + zstri * zstep + epos
            return __read_continuous(file, pos, shape, dtype,
                                     sequentialShape=seqShp)

    elif var.storage in ['99', '0', '00', '000', '1', '11', '111']:
    # elif var.storage == '99' or var.storage == '0':
        if tstep is None and zstep is None:
            shape = (1, var.zcount, var.ycount, var.xcount)
            if sequentialSize != -1:
                seqShp = (1, var.zcount, sequentialSize)
            else:
                seqShp = shape
            pos   = var.strPos + epos
            data  = []

            for l in range(var.tcount):
                data.append(__read_continuous(file, pos, shape, dtype,
                                              sequentialShape=seqShp))
                pos += tstride

            return np.concatenate(data)

        elif zstep is None and tstep is not None:
            shape = (1, var.zcount, var.ycount, var.xcount)
            if sequentialSize != -1:
                seqShp = (1, var.zcount, sequentialSize)
            else:
                seqShp = shape
            pos   = var.strPos + tstride * tstep + epos
            data = __read_continuous(file, pos, shape, dtype,
                                     sequentialShape=seqShp)

            return data

        elif tstep is None and zstep is not None:
            raise Exception('not implemented in 0,99')

        else:
            shape = (1, 1, var.ycount, var.xcount)
            zstri = var.ycount * var.xcount * 4
            if sequentialSize != -1:
                seqShp = (1, 1, sequentialSize)
                zstri += 8
            else:
                seqShp = shape
            pos   = var.strPos + tstride * tstep + zstri * zstep + epos
            return __read_continuous(file, pos, shape, dtype,
                                     sequentialShape=seqShp)

    else:
        raise Exception('invalid storage ' + var.storage +
                        ', only "99" or "-1,20" are supported')


def __read_continuous(file, offset=0, shape=None, dtype='<f4',
                      use_mmap=True, sequentialShape=None):
    """
    Read a block of continuous data into the memory.

    Parameters
    ----------
    file  : str
        A file from which data are read.
    offset: int
        An offset where the read is started.
    shape : tuple
        A tuple indicate shape of the Array returned.
    sequentialShape : tuple
        If in Fortran-sequential storage, provide a shape including the
        beginning and the end numbers.

    Returns
    -------
    re: numpy.ndarray
        Binary data.
    """
    with open(file, 'rb') as f:
        if use_mmap:
            data = np.memmap(f, dtype=dtype, mode='r', offset=offset,
                             shape=sequentialShape, order='C')
        else:
            number_of_values = reduce(lambda x, y: x*y, sequentialShape)

            f.seek(offset)
            data = np.fromfile(f, dtype=dtype, count=number_of_values)

    if sequentialShape != shape:
        data = data.reshape((shape[0],shape[1],-1))[:,:,1:-1]

    data = data.reshape(shape, order='C')

    data.shape = shape

    return data



# -*- coding: utf-8 -*-
"""
Created on 2020.08.01

@author: MiniUFO
Copyright 2018. All rights reserved. Use is subject to license terms.
"""
import xarray as xr
import numpy as np
import numba as nb
from numba.typed import List

_Rearth = 6371200


"""
Some map projection function useful for plotting
"""
def get_data_projection(ctl, globe=None, Rearth=_Rearth):
    """Get Projection
    
    Return the data projection indicated in PDEF for plot using cartopy.
    
    Parameters
    ----------
    ctl: str or CtlDescriptor
        Either a string representing a `ctl` file or a CtlDescriptor.
    globe: ccrs.Globe
        Default Globe parameter (None) in cartopy uses ellipse=WGS84.  Some
        regional numerical model like WRF use a spherical earth with a radius
        of 6370 km.  If one want to plot the data output from WRF with PDEF,
        one uses a default spherical globe with a radius of 6370000 and then
        provided this globe to this function for an accurate plot.
        
        Thanks to singledoggy at
        https://github.com/miniufo/xgrads/issues/32
        https://github.com/miniufo/xgrads/issues/37
    
    Returns
    -------
    re: cartopy.ccrs
        Cartopy coordinate reference system object for plotting.
    """
    import cartopy.crs as ccrs
    
    pdef = ctl.pdef
    
    if globe == None: # default globe
        globe = ccrs.Globe(ellipse='sphere',
                           semimajor_axis=Rearth, semiminor_axis=Rearth)
    
    if pdef is None:
        return ccrs.PlateCarree(globe=globe)
    else:
        PROJ = pdef.proj
        
        if PROJ is None:
            return ccrs.PlateCarree(globe=globe)
        
        elif PROJ in ['lcc', 'lccr']:
            from pyproj import Transformer
            
            if 'MOAD_CEN_LAT' in ctl.comments:
                clat = float(ctl.comments['MOAD_CEN_LAT'])
                
                tmp_proj = ccrs.LambertConformal(
                    globe=globe,
                    central_latitude=clat,
                    central_longitude=pdef.slon,
                    standard_parallels=(pdef.Struelat, pdef.Ntruelat)
                ).proj4_init
                
                transformer = Transformer.from_crs('WGS84', tmp_proj)
                e, n = transformer.transform(pdef.latref, pdef.lonref, errcheck=True)
                
                return ccrs.LambertConformal(globe=globe,
                          central_latitude   = clat,
                          central_longitude  = pdef.slon,
                          standard_parallels = (pdef.Struelat, pdef.Ntruelat),
                          false_easting  = (pdef.iref - 1) * pdef.dx - e,
                          false_northing = (pdef.jref - 1) * pdef.dy - n)
            else:
                clat = pdef.latref
                
                return ccrs.LambertConformal(globe=globe,
                          central_latitude   = clat,
                          central_longitude  = pdef.slon,
                          standard_parallels = (pdef.Struelat, pdef.Ntruelat),
                          false_easting  = pdef.iref * pdef.dx,
                          false_northing = pdef.jref * pdef.dy)
            
        elif PROJ == 'nps':
            return ccrs.NorthPolarStereo(globe=globe,
                      central_longitude = pdef.lonref,
                      true_scale_latitude = 60) # used by GrADS?
        
        elif PROJ == 'sps':
            return ccrs.SouthPolarStereo(globe=globe,
                      central_longitude = pdef.lonref,
                      true_scale_latitude = -60) # used by GrADS?


def interp_to_latlon(var, ctl):
    """Interpolate PDEF data onto lat/lon grids
    
    Interpolate the preprojected data onto lat/lon grids defined by ctl.
    
    Parameters
    ----------
    ctl: str or CtlDescriptor
        Either a string representing a `ctl` file or a CtlDescriptor.
    var: DataArray
        A variable defined at preprojected coordinates and
        need to be interpolated onto lat/lon grids.
    
    Returns
    -------
    re: xarray.DataArray
        Interpolated variable
    """
    if isinstance(ctl, str):
        from .core import CtlDescriptor
        ctl = CtlDescriptor(file=ctl)
    
    # get preprojected coordinates
    ypos, xpos = get_coordinates_from_PDEF(ctl, latlon=False)
    
    return var.interp(dict(y=ypos, x=xpos))


def get_coordinates_from_PDEF(ctl, latlon=True, Rearth=_Rearth):
    """
    Calculate coordinates based on the PDEF information.
    
    Parameters
    ----------
    ctl: str or CtlDescriptor
        Either a string representing a `ctl` file or a CtlDescriptor.
    latlon: boolean
        Return lat/lon coordinates on PDEF grids if True, from PDEF grids;
        Return PDEF coordinates if False, from lat/lon grids.
    Rearth: float
        Radius of the earth, default is 6371200m, consistent with GrADS.
        But this function is usually used in WRF data, then one should use
        6370000m, consistent with WRF model.
        Thanks to singledoggy at
        https://github.com/miniufo/xgrads/issues/32
        https://github.com/miniufo/xgrads/issues/37
    
    Returns
    -------
    y, x: xarray.DataArray, xarray.DataArray
        A tuple of (y, x) or (lat, lon)
    """
    if isinstance(ctl, str):
        from .core import CtlDescriptor
        ctl = CtlDescriptor(file=ctl)
    
    pdef = ctl.pdef
    
    if pdef is None:
        lats, lons = xr.broadcast(ctl.ydef.samples, ctl.xdef.samples)
        
        return lats, lons
    
    else:
        from pyproj import Proj #, transform
        
        PROJ = pdef.proj
        
        if   PROJ is None:
            lats, lons = xr.broadcast(ctl.ydef.samples, ctl.xdef.samples)
            
            return lats, lons
        
        elif PROJ in ['lcc', 'lccr']:
            # p = Proj(proj='lcc', datum='WGS84',
            #          lat_1=pdef.Struelat,
            #          lat_2=pdef.Ntruelat,
            #          lat_0=clat,
            #          lon_0=pdef.slon,
            #          x_0=pdef.iref * pdef.dx,
            #          y_0=pdef.jref * pdef.dy,
            #          a=_Rearth, b=_Rearth)
            p = Proj(get_data_projection(ctl, Rearth=Rearth).proj4_init)
            
            ypos = np.linspace(0, (pdef.jsize-1) * pdef.dx, pdef.jsize)
            xpos = np.linspace(0, (pdef.isize-1) * pdef.dy, pdef.isize)
        
        elif PROJ == 'nps':
            inc = pdef.gridinc * 1000 # change unit from km to m
            
            p = Proj(proj='stere',
                     lon_0=pdef.lonref, lat_0=90, lat_ts=60,
                     x_0=0, y_0=0,
                     a=_Rearth, b=_Rearth)
            
            ypos = np.linspace(-(pdef.jpole), (pdef.jsize-pdef.jpole),
                                 pdef.jsize) * inc
            xpos = np.linspace(-(pdef.ipole), (pdef.isize-pdef.ipole),
                                 pdef.isize) * inc
        
        elif PROJ == 'sps':
            inc = pdef.gridinc * 1000 # change unit from km to m
            
            p = Proj(proj='stere',
                     lon_0=pdef.lonref, lat_0=-90, lat_ts=-60,
                     x_0=0, y_0=0,
                     a=_Rearth, b=_Rearth)
            
            ypos = np.linspace(-(pdef.jpole), (pdef.jsize-pdef.jpole),
                                 pdef.jsize) * inc
            xpos = np.linspace(-(pdef.ipole), (pdef.isize-pdef.ipole),
                                 pdef.isize) * inc
        else:
            raise Exception('unsupported projection ' + PROJ)
        
        if latlon:
            xmesh, ymesh = np.meshgrid(xpos, ypos)
            
            reX, reY = p(xmesh, ymesh, inverse=True)
            
            reX = xr.DataArray(reX, dims=['y', 'x'],
                               coords={'x':xpos, 'y':ypos})
            reY = xr.DataArray(reY, dims=['y', 'x'],
                               coords={'x':xpos, 'y':ypos})
            
        else:
            lats, lons = ctl.ydef.samples, ctl.xdef.samples
            xmesh, ymesh = np.meshgrid(lons, lats)
            
            reX, reY = p(xmesh, ymesh, inverse=False)
            
            reX = xr.DataArray(reX, dims=['lat', 'lon'],
                               coords={'lat':lats, 'lon':lons})
            reY = xr.DataArray(reY, dims=['lat', 'lon'],
                               coords={'lat':lats, 'lon':lons})
        
        return reY, reX


def oacressman(dataS, lonS, latS, dimS, lonG, latG,
               rads=[10, 7, 4, 2, 1], undef=-9.99e8, method='cressman'):
    r"""Objective analysis using Cressmen method [1]_
    
    The function tries to reproduce `oacres` in GrADS.  Note that:
    
    1. The oacres function can be quite slow to execute, depending on grid
       and station data density;
    2. The scaling of the grid must be linear in lat-lon;
    3. The Cressman Analysis scheme can be unstable if the grid density is
       substantially higher than the station data density (ie, far more
       grid points than station data points). In such cases, the analysis
       can produce extrema in the grid values that are not realistic. It
       is thus suggested that you examine the results of oacres and compare
       them to the station data to insure they meet your needs.
    
    
    .. [1] Cressmen G. P., 1959: An operational objective analysis system,
       Monthly Weather Review, 87, 367–374.
    
    
    Parameters
    ----------
    dataS: DataArray
        A 1D DataArray representing the station data.
    lonS: DataArray
        A 1D array for longitudes of the stations.
    latS: DataArray
        A 1D array for latitudes of the stations.
    dimS: str
        Name of the station dimension.
    lonG: DataArray or numpy.array or list
        A 1D array for longitudes of the grid (should be linear).
    latG: DataArray or numpy.array or list
        A 1D array for latitudes of the grid (should be linear).
    rads: float or list of floats
        Radii at which the analysis is performed.  Default rads are the same
        as those of GrADS.
    undef: float
        Undefined value is set if a grid has no station nearby.
    method: str
        Methods of calculating weights, should one of ['cressman', 'exp'].
    
    Returns
    -------
    dataGrid: xarray.DataArray
        Result of the objective analysis.
    wei: xarray.DataArray
        Normalized total weights.
    """
    if isinstance(lonG, list):
        lonG = xr.DataArray(lonG, dims='lon', coords={'lon': lonG})
    if isinstance(latG, list):
        latG = xr.DataArray(latG, dims='lat', coords={'lat': latG})
    
    lonS = np.deg2rad(lonS)
    latS = np.deg2rad(latS)
    lonG = np.deg2rad(lonG)
    latG = np.deg2rad(latG)
    
    meanS = dataS.mean(dimS)
    dataG = latG  + lonG  # allocate one 2D slice
    dataG = dataG - dataG + meanS # initialize to mean value
    weisG = dataG - dataG # all set zeros
    
    radRs = ((latG[-1] - latG[0]) / (len(latG) - 1)).values * np.array(rads)
    
    tmp1, tmp2 = dataG, weisG
    for rad, RR in zip(radRs, rads):
        print(f'processing {np.rad2deg(rad):4.1f}-deg ({RR} grids) radius...')
        tmp1, tmp2 = xr.apply_ufunc(__slice, dataS, lonS, latS, lonG, latG,
                                    tmp1, tmp2, rad, undef,
                                    kwargs={'method':method},
                                    dask='allow',
                                    input_core_dims=[[dimS], [dimS], [dimS],
                                                     ['lon'], ['lat'],
                                                     ['lat','lon'],
                                                     ['lat','lon'], [], []],
                                    vectorize=True,
                                    output_core_dims=[['lat','lon'],
                                                      ['lat','lon']])
    
    re = tmp1.where(tmp1!=meanS) / len(radRs)
    re = re.where(~np.isnan(re), undef)
    
    return re.rename(dataG.name), (tmp2/tmp2.max()).rename('wei')


"""
Helper (private) methods are defined below
"""
def __slice(dataS, lonS, latS, lonG, latG, dataG, weis,
            rad, undef, method='cressman'):
    #print(dataS.shape, lonS.shape, latS.shape, lonG.shape,
    #      latG.shape, dataG.shape, rad, undef)
    if method == 'cressman':
        func_wei = __weight_cressman
    else:
        func_wei = __weight_exp
    
    stntag, stnwei, grdtag, grdwei = __cWeights(dataS, lonS, latS, lonG, latG,
                                                rad, func_wei)
    
    tmp = __interp_to_stations(dataG, stntag, stnwei, dataS, undef)
    
    return __interp_to_grids(dataS-tmp, grdtag, grdwei, dataG, weis, rad, undef)


@nb.jit(nopython=True, cache=False)
def __cWeights(dataS, lonS, latS, lonG, latG, rad, func_wei):
    r"""calculate weights and store them
    
    Parameters
    ----------
    dataS: numpy.ndarray
        A 1D DataArray representing the station data.
    lonS: numpy.ndarray
        A 1D array for longitudes of the stations.
    latS: numpy.ndarray
        A 1D array for latitudes of the stations.
    lonG: numpy.ndarray
        A 1D array for longitudes of the grid (should be linear).
    latG: numpy.ndarray
        A 1D array for latitudes of the grid (should be linear).
    rad: float
        Radius (in radian) at which the analysis is performed.
    
    Returns
    -------
    stntag: list
        Indices of grid points near a station
    stnwei: list
        Weights of grid points near a station
    grdtag: list
        Indices of stations near a grid point
    grdwei: list
        Weights of stations near a grid point
    """
    yc, xc = len(latG), len(lonG)
    sc     = len(dataS)
    
    stntag = List() # station indices within the radius
    stnwei = List() # weight coefficient of the stations
    grdtag = List() # grid indices within the radius
    grdwei = List() # weight coefficient of the grids
    
    # find grids
    for s in range(sc):
        tmp1 = [[0, 0]] # with a default one for numba to determine the type
        tmp2 = [np.nan]
        ########## find grid points near to a station ##########
        for j in range(yc):
            for i in range(xc):
                if np.abs(lonS[s]-lonG[i])<=rad*2 and \
                    np.abs(latS[s]-latG[j])<=rad:
                    sdis = __geodist(lonG[i], lonS[s], latG[j], latS[s])
                    
                    if sdis < rad:
                        rad2 = rad  ** 2.0
                        dis2 = sdis ** 2.0
                        tmp1.append([j, i])
                        tmp2.append(func_wei(rad2, dis2))
        
        stntag.append(np.array(tmp1))
        stnwei.append(np.array(tmp2))
    
    # find stations
    for j in range(yc):
        for i in range(xc):
            tmp1 = [0]  # with a default one for numba to determine the type
            tmp2 = [np.nan]
            ########## find stations near to a grid point ##########
            for s in range(sc):
                if np.abs(lonS[s]-lonG[i])<=rad*2 and \
                    np.abs(latS[s]-latG[j])<=rad:
                    sdis = __geodist(lonG[i], lonS[s], latG[j], latS[s])
                    
                    if sdis < rad:
                        rad2 = rad  ** 2.0
                        dis2 = sdis ** 2.0
                        tmp1.append(s)
                        tmp2.append(func_wei(rad2, dis2))
            
            grdtag.append(np.array(tmp1))
            grdwei.append(np.array(tmp2))
    
    return stntag, stnwei, grdtag, grdwei


@nb.jit(nopython=True, cache=False)
def __interp_to_stations(dataG, stntag, stnwei, dataS, undef=-9.99e8):
    x = dataS.shape[0]
    
    re = np.zeros(dataS.shape)
    
    for i in range(x):
        sum, wsum = 0, 0
        gc = len(stntag[i])
        
        if gc > 1:
            tags = stntag[i]
            weis = stnwei[i]
            for m in range(1, tags.shape[0]): # skip first value
                w = weis[m]
                jj, ii = tags[m]

                if dataG[jj, ii] != undef:
                    sum  += w * dataG[jj, ii]
                    wsum += w
            if wsum != 0:
                re[i] = sum / wsum
            else:
                re[i] = undef
        else:
            re[i] = undef
    
    return re


@nb.jit(nopython=True, cache=False)
def __interp_to_grids(dataS, grdtag, grdwei, dataG, weisG, rad, undef=-9.99e8):
    y, x = dataG.shape

    re = dataG.copy()
    we = weisG.copy()
    
    for j in range(y):
        for i in range(x):
            sum, wsum = 0, 0
            idx = j * x + i
            sc  = len(grdtag[idx])
            
            if sc > 1:
                tags = grdtag[idx]
                weis = grdwei[idx]
                for m in range(1, tags.shape[0]): # skip first value
                    w = weis[m]
                    t = tags[m]

                    if dataS[t] != undef:
                        sum  += w * dataS[t]
                        wsum += w
                
                if wsum != 0:
                    we[j,i] += wsum
                    if re[j,i] != undef:
                        re[j,i] += sum / wsum
                    else:
                        re[j,i] = sum / wsum
    
    return re, we


@nb.jit(nopython=True, cache=False)
def __geodist(lon1, lon2, lat1, lat2):
    """Calculate great-circle distance on a sphere of radius 1.
    
    Parameters
    ----------
    lon1: float
        Longitude for point 1 in radian.
    lon2: float
        Longitude for point 2 in radian.
    lat1: float
        Latitude  for point 1 in radian.
    lat2: float
        Latitude  for point 2 in radian.
    
    Returns
    -------
    dis: float
        Great circle distance (in radian)
    """
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    
    a = np.sin(dlat/2.0)**2.0 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2)**2.0
    
    dis = 2.0 * np.arcsin(np.sqrt(a))
    
    return dis


@nb.jit(nopython=True, cache=False)
def __weight_cressman(rad2, dis2):
    return (rad2-dis2)/(rad2+dis2)


@nb.jit(nopython=True, cache=False)
def __weight_exp(rad2, dis2):
    return np.exp(-dis2/(2.0*rad2))





# -*- coding: utf-8 -*-
from .core import CtlDescriptor
from .io import open_CtlDataset, open_mfdataset
from .utils import interp_to_latlon, get_coordinates_from_PDEF, \
                   get_data_projection, oacressman

__version__ = "0.2.6"



# -*- coding: utf-8 -*-
"""
Created on 2022.05.06

@author: MiniUFO
Copyright 2018. All rights reserved. Use is subject to license terms.
"""
#%%
import numpy as np
from pyproj import CRS
from xgrads import open_CtlDataset, get_data_projection,\
                          get_coordinates_from_PDEF

def test_llc_proj_wrf():
    # load data
    ctl_path = './ctls/grid.d1.ctl'
    dset, ctl = open_CtlDataset(ctl_path, returnctl=True)
    
    Rearth = 6370000 # consistent with WRF
    
    # get lat/lon and see the difference w.r.t WRF output
    lats, lons = get_coordinates_from_PDEF(ctl, Rearth=Rearth)

    XLAT, XLON = dset.XLAT.squeeze(), dset.XLONG.squeeze()
    
    assert np.isclose(np.abs(lats-XLAT), 0, atol=5e-5).all()
    assert np.isclose(np.abs(lons-XLON), 0, atol=5e-5).all()
    
    
    # write proj info into NetCDF
    crs = get_data_projection(ctl, Rearth=Rearth)
    projcrs = CRS(crs.proj4_init)
    
    cf = projcrs.to_cf()
    crs_new = projcrs.from_cf(cf)
    
    assert crs_new == projcrs





# -*- coding: utf-8 -*-
"""
Created on 2023.06.11

@author: MiniUFO
Copyright 2018. All rights reserved. Use is subject to license terms.
"""
#%% test objective analysis method
import numpy as np
import xarray as xr
from xgrads import oacressman


def test_oacressman():
    ds = xr.open_dataset('./ctls/stationPrecip.nc')
    
    lonV = np.linspace(70, 140, 141)
    latV = np.linspace(15,  55, 81)
    
    lon = xr.DataArray(lonV, dims='lon', coords={'lon': lonV})
    lat = xr.DataArray(latV, dims='lat', coords={'lat': latV})
    
    precip = xr.where(np.isnan(ds.precip), 0, ds.precip).load()
    
    res1, wei1 = oacressman(precip, ds.lons, ds.lats, 'stnID',
                           lon, lat, rads=[4], method='cressman')
    res2, wei2 = oacressman(precip, ds.lons, ds.lats, 'stnID',
                           lon, lat, rads=[4], method='exp')
    
    assert res1.shape == res2.shape == (31, 81, 141)
    assert wei1.shape == wei2.shape == (31, 81, 141)
    
    





# -*- coding: utf-8 -*-
"""
Created on 2020.03.02

@author: MiniUFO
Copyright 2018. All rights reserved. Use is subject to license terms.
"""
import numpy as np
import xarray as xr
from xgrads import open_CtlDataset, open_mfdataset


def test_template():
    dset1 = open_CtlDataset('./ctls/test8.ctl')
    dset2 = open_CtlDataset('./ctls/test9.ctl')
    dset3 = xr.tutorial.open_dataset('air_temperature')
    
    for l in range(len(dset1.time)):
        xr.testing.assert_equal(dset1.air[l], dset2.air[l])
        xr.testing.assert_equal(dset1.air[l], dset3.air[l])

    
    dset1 = open_mfdataset('./ctls/test8_*.ctl', parallel=True)
    dset2 = open_CtlDataset('./ctls/test8.ctl').load()
    dset3 = xr.tutorial.open_dataset('air_temperature').load()
    
    for l in range(len(dset1.time)):
        xr.testing.assert_equal(dset1.air[l], dset3.air[l])
    
    
    dset1 = open_mfdataset('./ctls/test9_*.ctl', parallel=True)
    dset2 = open_CtlDataset('./ctls/test9.ctl').load()
    dset3 = xr.tutorial.open_dataset('air_temperature').load()
    
    for l in range(len(dset1.time)):
        xr.testing.assert_equal(dset1.air[l], dset2.air[l])
    
    
    # test blank line in ctls
    dset1 = open_CtlDataset('./ctls/test81.ctl')
    dset2 = open_CtlDataset('./ctls/test82.ctl')
    
    assert (dset1.x == dset2.x).all()
    assert (dset1.y == dset2.y).all()
    assert (dset1.air[0] == dset2.air).all()

def test_ensemble():
    dset1 = open_CtlDataset('./ctls/ecmf_medium_T2m1.ctl')
    
    expected = np.array([2.011963 , 1.1813354, 1.1660767])
    
    # check several ensemble values
    assert np.isclose(dset1.t2[:,-1,-1,-1], expected).all()



# -*- coding: utf-8 -*-
"""
Created on 2020.03.02

@author: MiniUFO
Copyright 2018. All rights reserved. Use is subject to license terms.
"""
#%%
import numpy as np
import pytest
from xgrads import CtlDescriptor


def test_parse_str():
    content=\
        "dset ^Model10\n"\
        "*ccccccccccccccccc\n"\
        "title 10-deg resolution model\n"\
        "undef -9.99e8\n"\
        "xdef 36 linear   0 10\n"\
        "ydef 19 linear -90 10\n"\
        "zdef   1 linear 0 1\n"\
        "tdef   1 linear 00z01Jan2000 1dy\n"\
        "vars 1\n"\
        "test 1 99 test variable\n"\
        "endvars\n"
    
    ctl = CtlDescriptor(content=content)
    
    assert ctl.dsetPath == '^Model10'
    assert ctl.descPath == None
    assert ctl.title == '10-deg'
    assert ctl.undef == -9.99e8
    assert ctl.zrev == False
    assert ctl.yrev == False
    assert ctl.template == False
    assert ctl.periodicX == True
    assert ctl.cal365Days == False
    assert ctl.sequential == False
    assert ctl.byteOrder == 'little'
    assert (ctl.xdef.samples == np.linspace(0, 350, 36)).all()
    assert (ctl.ydef.samples == np.linspace(-90, 90, 19)).all()
    assert (ctl.zdef.samples == np.array([0.])).all()
    assert (ctl.tdef.samples == np.array(['2000-01-01T00:00:00'], dtype='datetime64[s]'))
    assert ctl.pdef == None


    content=\
        "dset D:/Data/ERAInterim/ERAWindAmplitudes.dat\n"\
        "undef -9999.0\n"\
        "title null\n"\
        "xdef 240 linear 0.0 1.5\n"\
        "ydef 121 linear -90.0 1.5\n"\
        "zdef 1 linear 1000.0 50.0\n"\
        "tdef 1 linear 00:00Z01JAN1993 24hr\n"\
        "vars 12\n"\
        "u10amp0 0 99 amplitude of harmonics of period 365.0\n"\
        "u10amp1 0 99 amplitude of harmonics of period 182.0\n"\
        "u10amp2 0 99 amplitude of harmonics of period 91.0\n"\
        "u10amp3 0 99 amplitude of harmonics of period 60.0\n"\
        "v10amp0 0 99 amplitude of harmonics of period 365.0\n"\
        "v10amp1 0 99 amplitude of harmonics of period 182.0\n"\
        "v10amp2 0 99 amplitude of harmonics of period 91.0\n"\
        "v10amp3 0 99 amplitude of harmonics of period 60.0\n"\
        "mamp0 0 99 amplitude of harmonics of period 365.0\n"\
        "mamp1 0 99 amplitude of harmonics of period 182.0\n"\
        "mamp2 0 99 amplitude of harmonics of period 91.0\n"\
        "mamp3 0 99 amplitude of harmonics of period 60.0\n"\
        "endvars\n"
    
    ctl = CtlDescriptor(content=content)
        
    assert ctl.dsetPath == 'D:/Data/ERAInterim/ERAWindAmplitudes.dat'
    assert ctl.descPath == None
    assert ctl.title == 'null'
    assert ctl.undef == -9999.0
    assert ctl.zrev == False
    assert ctl.yrev == False
    assert ctl.template == False
    assert ctl.periodicX == True
    assert ctl.cal365Days == False
    assert ctl.sequential == False
    assert ctl.byteOrder == 'little'
    assert (ctl.xdef.samples == np.linspace(0, 358.5, 240)).all()
    assert (ctl.ydef.samples == np.linspace(-90, 90, 121)).all()
    assert (ctl.zdef.samples == np.array([1000.])).all()
    assert (ctl.tdef.samples == np.array(['1993-01-01T00:00:00'], dtype='datetime64[s]'))
    assert ctl.pdef == None


def test_coding():
    with pytest.raises(UnicodeDecodeError):
        ctl = CtlDescriptor(file='./ctls/test0.ctl')
    
    ctl = CtlDescriptor(encoding='UTF-8', file='./ctls/test0.ctl')
    
    assert ctl.dsetPath == './ctls/intensity16070412.dat'


def test_all_sample_ctls():
    paths = [
        './ctls/19855pt.grb',
        './ctls/2017101712_1721.dat',
        './ctls/ElenaIsenTest.dat',
        './ctls/flux_stn.dat',
        './ctls/KeffUninterp.dat',
        './ctls/Shanshan.dat',
        './ctls/data.grads',
        ['./ctls/test8_2013010100.dat', './ctls/test8_2013010106.dat',
         './ctls/test8_2013010112.dat', './ctls/test8_2013010118.dat'],
        ['./ctls/test9_20130101.dat', './ctls/test9_20130102.dat',
         './ctls/test9_20130103.dat'],
        './ctls/test10.dat',
    ]
    for i in range(10):
        ctl = CtlDescriptor(file='./ctls/test'+str(i+1)+'.ctl')
        re = ctl.dsetPath == paths[i]
        
        if type(re) == bool:
            assert re
        else:
            assert re.all()


def test_ensemble_ctls():
    ctl1 = CtlDescriptor(file='./ctls/ecmf_medium_T2m1.ctl')
    
    assert len(ctl1.edef) == 3
    
    ens = ['c00', 'p01', 'p02']
    
    strPos = 0
    for name, en in zip(ens, ctl1.edef):
        assert name == en.name
        assert en.strPos == strPos
        strPos += en.tcount * ctl1.tRecLength
    
    ctl2 = CtlDescriptor(file='./ctls/ecmf_medium_T2m2.ctl')
    
    assert len(ctl2.edef) == 4
    
    enames = ['ensm', 'm01', 'm02', 'm03']
    tcount = [41, 40, 39, 38]
    codes  = [None, '3,1', None, '3,2']
    
    strPos = 0
    for name, tc, code, en in zip(enames, tcount, codes, ctl2.edef):
        assert name == en.name
        assert tc   == en.tcount
        assert code == en.codes
        assert en.strPos == strPos
        strPos += en.tcount * ctl1.tRecLength

